@@ -27,18 +27,20 @@
 
 #include "rendercheck.h"
 
+XRenderPictFormat **format_list;
+int nformats;
+int argb32index;
+
 /* Note: changing the order of these colors may disrupt tests that depend on
- * specific colors.  Just add to the end if you need.
+ * specific colors.  Just add to the end if you need.  These are
+ * not premultiplied, but will be in main().
  */
 color4d colors[] = {
 	{1.0, 1.0, 1.0, 1.0},
 	{1.0, 0, 0, 1.0},
 	{0, 1.0, 0, 1.0},
 	{0, 0, 1.0, 1.0},
-	{0.5, 0, 0, 0.25},
 	{0.5, 0, 0, .5},
-	{0.0, .5, 1.0, .5},
-	{0.0, .5, 1.0, 0}
 };
 
 /* Convenience pointers to 1x1 repeating colors */
@@ -154,7 +156,7 @@ eval_diff(char *name, color4d *expected, color4d *test, int x, int y,
 	rscale = 1.0 * (1 << 5);
 	gscale = 1.0 * (1 << 6);
 	bscale = 1.0 * (1 << 5);
-	ascale = 1.0;
+	ascale = 1.0 * 32;
 	rdiff = fabs(test->r - expected->r) * rscale;
 	bdiff = fabs(test->g - expected->g) * gscale;
 	gdiff = fabs(test->b - expected->b) * bscale;
@@ -166,6 +168,7 @@ eval_diff(char *name, color4d *expected, color4d *test, int x, int y,
 	diff = max(max(max(rdiff, gdiff), bdiff), adiff);
 	if (diff > 3.0) {
 		printf("%s test error of %.4f at (%d, %d) --\n"
+		    "           R    G    B    A\n"
 		    "got:       %.2f %.2f %.2f %.2f\n"
 		    "expected:  %.2f %.2f %.2f %.2f\n", name, diff, x, y,
 		    test->r, test->g, test->b, test->a,
@@ -193,28 +196,108 @@ argb_fill(Display *dpy, picture_info *p, int x, int y, int w, int h, float a,
 	XRenderFillRectangle(dpy, PictOpSrc, p->pict, &rendercolor, x, y, w, h);
 }
 
+#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
+
+/* Create a set of direct format XRenderPictFormats for later use.  This lets
+ * us get more formats than just the standard required set, and lets us attach
+ * names to them.
+ */
+void
+create_formats_list(Display *dpy)
+{
+    int i;
+    int nformats_allocated = 5;
+    XRenderPictFormat templ;
+
+    memset(&templ, 0, sizeof(templ));
+    templ.type = PictTypeDirect;
+
+    format_list = malloc(sizeof(XRenderPictFormat *) * nformats_allocated);
+    if (format_list == NULL)
+	errx(1, "malloc error");
+    nformats = 0;
+
+    argb32index = -1;
+    for (i = 0; ; i++) {
+	char name[20];
+	int alphabits, redbits;
+
+	if (nformats + 1 == nformats_allocated) {
+	    nformats_allocated *= 2;
+	    format_list = realloc(format_list, sizeof(XRenderPictFormat *) *
+		nformats_allocated);
+	    if (format_list == NULL)
+		errx(1, "realloc error");
+	}
+
+	format_list[nformats] = XRenderFindFormat(dpy, PictFormatType, &templ,
+	    i);
+	if (format_list[nformats] == NULL)
+	    break;
+
+	alphabits = bit_count(format_list[nformats]->direct.alphaMask);
+	redbits = bit_count(format_list[nformats]->direct.redMask);
+
+	/* Our testing code isn't all that hot, so don't bother trying at
+	 * the low depths yet.
+	 */
+	if ((redbits >= 1 && redbits <= 4) ||
+	    (alphabits >= 1 && alphabits <= 4))
+	{
+	    continue;
+	}
+
+	describe_format(name, 20, format_list[nformats]);
+
+	if (format_whitelist_len != 0) {
+	    Bool ok = FALSE;
+	    int j;
+
+	    for (j = 0; j < format_whitelist_len; j++) {
+		if (strcmp(format_whitelist[j], name) == 0) {
+		    ok = TRUE;
+		    break;
+		}
+	    }
+	    if (!ok) {
+		printf("Ignoring server-supported format: %s\n", name);
+		continue;
+	    }
+	}
+
+	if (format_list[nformats] == XRenderFindStandardFormat(dpy,
+	    PictStandardARGB32))
+	{
+	    argb32index = nformats;
+	}
+
+	printf("Found server-supported format: %s\n", name);
+
+	nformats++;
+    }
+    if (argb32index == -1) {
+	errx(1, "required ARGB32 format not found");
+    }
+}
+
 Bool
 do_tests(Display *dpy, picture_info *win)
 {
 	int i, j, src, dst = 0, mask;
-	int num_dests, num_formats;
+	int num_dests;
 	picture_info *dests, *pictures_1x1, *pictures_10x10, picture_3x3, *pictures_solid;
 	int success_mask = 0, tests_passed = 0, tests_total = 0;
+	int num_tests;
 
-	num_dests = 3;
+	create_formats_list(dpy);
+
+	num_dests = nformats;
 	dests = (picture_info *)malloc(num_dests * sizeof(dests[0]));
 	if (dests == NULL)
 		errx(1, "malloc error");
 
-	dests[0].format = XRenderFindStandardFormat(dpy, PictStandardARGB32);
-	dests[1].format = XRenderFindStandardFormat(dpy, PictStandardRGB24);
-	dests[2].format = XRenderFindStandardFormat(dpy, PictStandardA8);
-	/*
-	dests[3].format = XRenderFindStandardFormat(dpy, PictStandardA4);
-	dests[4].format = XRenderFindStandardFormat(dpy, PictStandardA1);
-	*/
-
 	for (i = 0; i < num_dests; i++) {
+		dests[i].format = format_list[i];
 		dests[i].d = XCreatePixmap(dpy, RootWindow(dpy, 0),
 		    win_width, win_height, dests[i].format->depth);
 		dests[i].pict = XRenderCreatePicture(dpy, dests[i].d,
@@ -226,20 +309,17 @@ do_tests(Display *dpy, picture_info *win)
 		describe_format(dests[i].name, 20, dests[i].format);
 	}
 
-	num_formats = 3;
-
-	pictures_1x1 = (picture_info *)malloc(num_colors * num_formats *
+	pictures_1x1 = (picture_info *)malloc(num_colors * nformats *
 	    sizeof(picture_info));
 	if (pictures_1x1 == NULL)
 		errx(1, "malloc error");
 
-	for (i = 0; i < num_colors * num_formats; i++) {
+	for (i = 0; i < num_colors * nformats; i++) {
 		XRenderPictureAttributes pa;
-		color4d *c = &colors[i / num_formats];
+		color4d *c = &colors[i / nformats];
 
 		/* The standard PictFormat numbers go from 0 to 4 */
-		pictures_1x1[i].format = XRenderFindStandardFormat(dpy,
-		    i % num_formats);
+		pictures_1x1[i].format = format_list[i % nformats];
 		pictures_1x1[i].d = XCreatePixmap(dpy, RootWindow(dpy, 0), 1,
 		    1, pictures_1x1[i].format->depth);
 		pa.repeat = TRUE;
@@ -260,23 +340,22 @@ do_tests(Display *dpy, picture_info *win)
 		pictures_1x1[i].color = *c;
 		color_correct(&pictures_1x1[i], &pictures_1x1[i].color);
 	}
-	argb32white = &pictures_1x1[0 * num_formats];
-	argb32red = &pictures_1x1[1 * num_formats];
-	argb32green = &pictures_1x1[2 * num_formats];
-	argb32blue = &pictures_1x1[3 * num_formats];
+	argb32white = &pictures_1x1[0 * nformats + argb32index];
+	argb32red = &pictures_1x1[1 * nformats + argb32index];
+	argb32green = &pictures_1x1[2 * nformats + argb32index];
+	argb32blue = &pictures_1x1[3 * nformats + argb32index];
 
-	pictures_10x10 = (picture_info *)malloc(num_colors * num_formats *
+	pictures_10x10 = (picture_info *)malloc(num_colors * nformats *
 	    sizeof(picture_info));
 	if (pictures_10x10 == NULL)
 		errx(1, "malloc error");
 
-	for (i = 0; i < num_colors * num_formats; i++) {
+	for (i = 0; i < num_colors * nformats; i++) {
 		XRenderPictureAttributes pa;
-		color4d *c = &colors[i / num_formats];
+		color4d *c = &colors[i / nformats];
 
 		/* The standard PictFormat numbers go from 0 to 4 */
-		pictures_10x10[i].format = XRenderFindStandardFormat(dpy,
-		    i % num_formats);
+		pictures_10x10[i].format = format_list[i % nformats];
 		pictures_10x10[i].d = XCreatePixmap(dpy, RootWindow(dpy, 0), 10,
 		    10, pictures_10x10[i].format->depth);
 		pa.repeat = TRUE;
@@ -320,6 +399,7 @@ do_tests(Display *dpy, picture_info *win)
             c.green = (int)(colors[i].g*65535);
             c.blue = (int)(colors[i].b*65535);
             pictures_solid[i].pict = XRenderCreateSolidFill(dpy, &c);
+            pictures_solid[i].format = format_list[argb32index];
             pictures_solid[i].name = "Solid";
         }
 
@@ -331,17 +411,19 @@ do {								\
 	tests_total++;						\
 } while (0)
 
+	num_tests = num_colors * nformats;
+
 	if (enabled_tests & TEST_FILL) {
 		Bool ok, group_ok = TRUE;
 
 		printf("Beginning testing of filling of 1x1R pictures\n");
-		for (i = 0; i < num_colors * num_formats; i++) {
+		for (i = 0; i < num_tests; i++) {
 			ok = fill_test(dpy, win, &pictures_1x1[i]);
 			RECORD_RESULTS();
 		}
 
 		printf("Beginning testing of filling of 10x10 pictures\n");
-		for (i = 0; i < num_colors * num_formats; i++) {
+		for (i = 0; i < num_tests; i++) {
 			ok = fill_test(dpy, win, &pictures_10x10[i]);
 			RECORD_RESULTS();
 		}
@@ -353,9 +435,8 @@ do {								\
 		Bool ok, group_ok = TRUE;
 
 		printf("Beginning dest coords test\n");
-		/* 0 and num_formats should result in ARGB8888 red on ARGB8888 white. */
-		ok = dstcoords_test(dpy, win, &dests[0], &pictures_1x1[0],
-		    &pictures_1x1[num_formats]);
+		ok = dstcoords_test(dpy, win, &dests[0], argb32white,
+		    argb32red);
 		RECORD_RESULTS();
 		if (group_ok)
 			success_mask |= TEST_DSTCOORDS;
@@ -365,7 +446,7 @@ do {								\
 		Bool ok, group_ok = TRUE;
 
 		printf("Beginning src coords test\n");
-		ok = srccoords_test(dpy, win, &pictures_1x1[0], FALSE);
+		ok = srccoords_test(dpy, win, argb32white, FALSE);
 		RECORD_RESULTS();
 		if (group_ok)
 			success_mask |= TEST_SRCCOORDS;
@@ -375,7 +456,7 @@ do {								\
 		Bool ok, group_ok = TRUE;
 
 		printf("Beginning mask coords test\n");
-		ok = srccoords_test(dpy, win, &pictures_1x1[0], TRUE);
+		ok = srccoords_test(dpy, win, argb32white, TRUE);
 		RECORD_RESULTS();
 		if (group_ok)
 			success_mask |= TEST_MASKCOORDS;
@@ -385,7 +466,7 @@ do {								\
 		Bool ok, group_ok = TRUE;
 
 		printf("Beginning transformed src coords test\n");
-		ok = trans_coords_test(dpy, win, &pictures_1x1[0], FALSE);
+		ok = trans_coords_test(dpy, win, argb32white, FALSE);
 		RECORD_RESULTS();
 		if (group_ok)
 			success_mask |= TEST_TSRCCOORDS;
@@ -395,7 +476,7 @@ do {								\
 		Bool ok, group_ok = TRUE;
 
 		printf("Beginning transformed mask coords test\n");
-		ok = trans_coords_test(dpy, win, &pictures_1x1[0], TRUE);
+		ok = trans_coords_test(dpy, win, argb32white, TRUE);
 		RECORD_RESULTS();
 		if (group_ok)
 			success_mask |= TEST_TMASKCOORDS;
@@ -405,6 +486,9 @@ do {								\
 		Bool ok, group_ok = TRUE;
 
 		for (i = 0; i < num_ops; i++) {
+		    if (ops[i].disabled)
+			continue;
+
 		    for (j = 0; j <= num_dests; j++) {
 			picture_info *pi;
 
@@ -415,8 +499,8 @@ do {								\
 			printf("Beginning %s blend test on %s\n", ops[i].name,
 			    pi->name);
 
-			for (src = 0; src < num_colors * num_formats; src++) {
-				for (dst = 0; dst < num_colors; dst++) {
+			for (src = 0; src < num_tests; src++) {
+				for (dst = 0; dst < num_tests; dst++) {
 					ok = blend_test(dpy, win, pi, i,
 					    &pictures_1x1[src],
 					    &pictures_1x1[dst]);
@@ -443,6 +527,9 @@ do {								\
 		Bool ok, group_ok = TRUE;
 
 		for (i = 0; i < num_ops; i++) {
+		    if (ops[i].disabled)
+			continue;
+
 		    for (j = 0; j <= num_dests; j++) {
 			picture_info *pi;
 
@@ -453,9 +540,9 @@ do {								\
 			printf("Beginning %s composite mask test on %s\n",
 			    ops[i].name, pi->name);
 
-			for (src = 0; src < num_colors; src++) {
-			    for (mask = 0; mask < num_colors; mask++) {
-				for (dst = 0; dst < num_colors; dst++) {
+			for (src = 0; src < num_tests; src++) {
+			    for (mask = 0; mask < num_tests; mask++) {
+				for (dst = 0; dst < num_tests; dst++) {
 					ok = composite_test(dpy, win, pi, i,
 					    &pictures_10x10[src],
 					    &pictures_10x10[mask],
@@ -489,6 +576,9 @@ do {								\
 		Bool ok, group_ok = TRUE;
 
 		for (i = 0; i < num_ops; i++) {
+		    if (ops[i].disabled)
+			continue;
+
 		    for (j = 0; j <= num_dests; j++) {
 			picture_info *pi;
 
@@ -499,9 +589,9 @@ do {								\
 			printf("Beginning %s composite CA mask test on %s\n",
 			    ops[i].name, pi->name);
 
-			for (src = 0; src < num_colors; src++) {
-			    for (mask = 0; mask < num_colors; mask++) {
-				for (dst = 0; dst < num_colors; dst++) {
+			for (src = 0; src < num_tests; src++) {
+			    for (mask = 0; mask < num_tests; mask++) {
+				for (dst = 0; dst < num_tests; dst++) {
 					ok = composite_test(dpy, win, pi, i,
 					    &pictures_10x10[src],
 					    &pictures_10x10[mask],
@@ -535,6 +625,9 @@ do {								\
 	    Bool ok, group_ok = TRUE;
 
             for (i = 0; i < num_ops; i++) {
+		if (ops[i].disabled)
+		    continue;
+
                 for (j = 0; j <= num_dests; j++) {
                     picture_info *pi;
                     
@@ -545,8 +638,8 @@ do {								\
                     printf("Beginning %s linear gradient test on %s\n",
                            ops[i].name, pi->name);
                     
-                    for (src = 0; src < num_colors; src++) {
-                        for (mask = 0; mask < num_colors; mask++) {
+                    for (src = 0; src < num_tests; src++) {
+                        for (mask = 0; mask < num_tests; mask++) {
                             ok = linear_gradient_test(dpy, win, pi, i,
 				&pictures_1x1[dst]);
 			    RECORD_RESULTS();
@@ -562,6 +655,9 @@ do {								\
 	    Bool ok, group_ok = TRUE;
 
             for (i = 0; i < num_ops; i++) {
+		if (ops[i].disabled)
+		    continue;
+
                 for (j = 0; j <= num_dests; j++) {
                     picture_info *pi;
                     
@@ -596,6 +692,9 @@ do {								\
 	    Bool ok, group_ok = TRUE;
 
 	    for (i = 0; i < num_ops; i++) {
+		if (ops[i].disabled)
+		    continue;
+
 		for (j = 0; j <= num_dests; j++) {
 			picture_info *pi;
 
@@ -607,19 +706,19 @@ do {								\
 			printf("Beginning %s Triangles test on %s\n",
 			    ops[i].name, pi->name);
 			ok = triangles_test(dpy, win, pi, i,
-			    &pictures_1x1[num_formats], &pictures_1x1[0]);
+			    argb32red, argb32white);
 			RECORD_RESULTS();
 
 			printf("Beginning %s TriStrip test on %s\n",
 			    ops[i].name, pi->name);
 			ok = tristrip_test(dpy, win, pi, i,
-			    &pictures_1x1[num_formats], &pictures_1x1[0]);
+			    argb32red, argb32white);
 			RECORD_RESULTS();
 
 			printf("Beginning %s TriFan test on %s\n",
 			    ops[i].name, pi->name);
 			ok = trifan_test(dpy, win, pi, i,
-			    &pictures_1x1[num_formats], &pictures_1x1[0]);
+			    argb32red, argb32white);
 			RECORD_RESULTS();
 		}
 	    }
@@ -627,7 +726,35 @@ do {								\
 		success_mask |= TEST_TRIANGLES;
 	}
 
+        if (enabled_tests & TEST_BUG7366) {
+	    Bool ok, group_ok = TRUE;
+
+	    ok = bug7366_test(dpy);
+	    RECORD_RESULTS();
+
+	    if (group_ok)
+		success_mask |= TEST_BUG7366;
+	}
+
 	printf("%d tests passed of %d total\n", tests_passed, tests_total);
 
 	return tests_passed == tests_total;
 }
+
+/**
+ * \brief copies the contents of a picture to the window.
+ *
+ * This is used in tests so that the user sees flashing colors indicating that
+ * rendercheck is really doing things.  The minimalrendering commandline option
+ * indicates that this behavior should be disabled.
+ */
+void
+copy_pict_to_win(Display *dpy, picture_info *pict, picture_info *win,
+    int width, int height)
+{
+	if (pict == win || minimalrendering)
+		return;
+
+	XRenderComposite(dpy, PictOpSrc, pict->pict, 0, win->pict, 0, 0,
+	    0, 0, 0, 0, width, height);
+}
